#include "xlistinfo.h"
#include "xlistobject.h"
#include "kernel.h"
#include "xquery.h"

#define DEFERRED_ADDRESS(ADDR) 0


static PyObject* empty_list;
static PyObject* empty_dict;

PyObject* kernel_strip_c(PyObject *seq);

static int
uriAndPrefix_fromtype(PyTypeObject* type, PyObject** uri, PyObject** prefix) {
	PyObject *moduleName, *sys_modules, *module;
	
	sys_modules = PySys_GetObject("modules");
	moduleName = PyObject_GetAttrString((PyObject*)type, "__module__");
	module = PyDict_GetItem(sys_modules, moduleName);
	if (module == NULL) {
		PyErr_SetObject(PyExc_AttributeError, 
			PyString_FromFormat("Unable to find module %s while initializing %s.",
				PyString_AsString(moduleName),
				type->tp_name));
		return -1;
	}

	*uri = PyObject_GetAttrString(module, "__uri__");
	if (*uri == NULL) {
		Satine_WriteInfo("Undefined __uri__ in module %s while initializing %s.\n",
			PyString_AsString(moduleName),
			type->tp_name);
		PyErr_SetObject(PyExc_AttributeError, 
			PyString_FromFormat("Undefined __uri__ in module %s while initializing %s. Did you call xspace?",
				PyString_AsString(moduleName),
				type->tp_name));
		return -1;
	}

	*prefix = PyObject_GetAttrString(module, "__prefix__");
	if (*prefix == NULL) {
		Satine_WriteInfo("Undefined __prefix__ in module %s while initializing %s.\n",
			PyString_AsString(moduleName),
			type->tp_name);
		PyErr_SetObject(PyExc_AttributeError, 
			PyString_FromFormat("Undefined __prefix__ in module %s while initializing %s. Did you call xspace?",
				PyString_AsString(moduleName),
				type->tp_name));
		return -1;
	}

	

	return 0;
}

static int 
xlistinfo_create_tag(PyXListInfo* xinfo, PyTypeObject* type) {
	xinfo->tag = PyDict_GetItemString(type->tp_dict, "__tag__");
	if (xinfo->tag == NULL)
		xinfo->tag = PyString_FromString(type->tp_name);
	else
		if (!PyString_Check(xinfo->tag)) {
			Satine_WriteInfo("In type %s, __tag___ field %s is not a valid identifier\n", type->tp_name,
				PyString_AsString(PyObject_Repr(xinfo->tag)));
			PyErr_SetString(PyExc_AttributeError, "__tag__ must be a string");
			return -1;
		}
	return 0;
}

static int 
xlistinfo_create_uriAndPrefix(PyXListInfo* xinfo, PyTypeObject* type) {
	xinfo->uri = PyDict_GetItemString(type->tp_dict, "__uri__");
	xinfo->prefix = PyDict_GetItemString(type->tp_dict, "__prefix__");
	if (xinfo->uri == NULL || xinfo->prefix == NULL){
			if (uriAndPrefix_fromtype(type, &xinfo->uri, &xinfo->prefix) == -1) 
				return -1;
		}
	return 0;
}


static int 
xlistinfo_compile_attrs(PyXListInfo* xinfo, PyTypeObject* type) {
	PyObject* attrs;
	PyObject* attrs_as_xlist;
	PyObject* key;
	PyObject* use;
	PyObject* value;
	PyObject* default_value;

	char* attrs__c;
	int attrs__l;
	int i;
	int size;
	struct v_node* node = NULL;

	Satine_WriteMoreInfo("compile attrs\n");

	attrs = PyDict_GetItemString(type->tp_dict, "__attrs__");


	if (attrs != Py_None && attrs != NULL) {
		xinfo->attrs_types = PyDict_New();
		xinfo->default_attrs = PyDict_New();
		xinfo->none_attrs = PyList_New(0);
		xinfo->required_attrs = PyList_New(0);

		if (PyString_Check(attrs)) {
			Satine_WriteMoreInfo("Compiling %s.__attrs__ = %s\n", type->tp_name,
				PyString_AsString(attrs));

			PyString_AsStringAndSize(attrs, &attrs__c, &attrs__l);
			attrs_as_xlist = kernel_xml2seq_c(attrs__c, attrs__l, NULL, 0);
			if (attrs_as_xlist == NULL) {
				Satine_WriteInfo("In type %s, __attrs___ field is not a valid XML fragment\n", type->tp_name);
				PyErr_SetString(PyExc_AttributeError, "Bad format for attributes definition (field __attrs__ in class definition)");
				return -1;
			}
		}
		else if (PyXList_Check(attrs)) {
			attrs_as_xlist = attrs;
		}
		else 
			return -1;

		if (attrs_as_xlist == NULL) {
			PyErr_SetString(PyExc_AttributeError, "Bad format for attribute __attrs__");
			return -1;
		}

		attrs_as_xlist = kernel_strip_c(attrs_as_xlist);
		if (attrs_as_xlist == NULL) {
			PyErr_SetString(PyExc_AttributeError, "Bad format for attribute __attrs__");
			return -1;
		}

		size = PyList_GET_SIZE(attrs_as_xlist);
		if (size/2*2 != size) {
			Satine_WriteInfo("Bad format for attributes definition for type %s\n", type->tp_name);
			PyErr_SetString(PyExc_AttributeError, "Bad format for attributes definition (field __attrs__ in class definition)");
			return -1;
		}

		Satine_WriteMoreInfo("Browsing %s attributes = %s\n", type->tp_name,
			PyString_AsString(PyObject_Repr(attrs_as_xlist)));
		for (i=0; i < size; ) {
			value = PyList_GetItem(attrs_as_xlist, i++);
			key = PyList_GetItem(attrs_as_xlist, i++);
			if (key == NULL)
				break;

			if (!PyObject_GetAttrString(value, "xml2py")) {
				PyErr_SetObject(PyExc_AttributeError, PyString_FromFormat( 
					"Attribute validation: object %s in attribute %s does not provide the method xml2py",
					PyString_AsString(PyObject_Repr(value)),
					PyString_AsString(key)));
				return -1;
			}
			if (!PyObject_GetAttrString(value, "py2xml")) {
				PyErr_SetObject(PyExc_AttributeError, PyString_FromFormat(
					"During validation, object %s in attribute %s does not provide the method py2xml",
					PyString_AsString(PyObject_Repr(value)),
					PyString_AsString(key)));
				return -1;
			}
			PyDict_SetItem(xinfo->attrs_types, key, value);
			default_value = PyObject_GetAttrString(value, "default");
			if (default_value) {
				PyDict_SetItem(xinfo->default_attrs, key, default_value);
			}

			PyList_Append(xinfo->none_attrs, key);

			use = PyObject_GetAttrString(value, "use");
			if (use) {
				if (PyObject_Compare(use, PyString_InternFromString("required")) == 0)
					PyList_Append(xinfo->required_attrs, key);
				Py_DECREF(use);
			}
		}
	}
	else {
		PyErr_Clear();
		Py_INCREF(Py_None);
		Py_INCREF(empty_dict);
		Py_INCREF(empty_list);
		Py_INCREF(empty_dict);

		xinfo->attrs_types = Py_None;
		xinfo->default_attrs = empty_dict;
		xinfo->required_attrs = empty_list;
		xinfo->none_attrs = empty_dict;
	}

	Satine_WriteMoreInfo("Validation init for %s\n\
		\tattrs_types = %s\n\
		\tdefault_attrs = %s\n\
		\trequires_attrs = %s\n\
		\tnone_attrs = %s\n", 
			type->tp_name,
			PyString_AsString(PyObject_Repr(xinfo->attrs_types)),
			PyString_AsString(PyObject_Repr(xinfo->default_attrs)),
			PyString_AsString(PyObject_Repr(xinfo->required_attrs)),
			PyString_AsString(PyObject_Repr(xinfo->none_attrs)));


	return 0;
}


static int 
xlistinfo_compile_items(PyXListInfo* xinfo, PyTypeObject* type) {
	PyObject* items;

	Satine_WriteMoreInfo("compile items\n");

	items = PyDict_GetItemString(type->tp_dict, "__items__");
	if (items != Py_None && items != NULL) {
		xinfo->items_dfa = xquery_compile(items);
		if (xinfo->items_dfa == NULL) {
			Satine_WriteInfo("unable to compile __items__ to a dfa for type %s\n", type->tp_name);
			PyErr_SetString(PyExc_AttributeError, "__items__ is not valid a XRE");
			return -1;
		}
	}
	else {
		PyErr_Clear();
		Py_INCREF(Py_None);
		xinfo->items_dfa = Py_None;
	}

	Satine_WriteMoreInfo("compile items end\n");
	return 0;
}

PyObject* PyXListInfo_GetDatatype(PyXListInfo* tpinfo, PyObject* key) {
	return PyDict_GetItem(tpinfo->attrs_types, key);
}

PyObject* 
PyXListInfo_ForXList() {
	PyXListInfo* xinfo;
	xinfo = (PyXListInfo*) PyObject_NEW(PyXListInfo, &PyXListInfo_Type);
	if (xinfo == NULL) {
		return NULL;
	}

	Py_INCREF(Py_None);
	Py_INCREF(Py_None);
	Py_INCREF(empty_dict);
	Py_INCREF(empty_list);
	Py_INCREF(empty_list);

	
	xinfo->tag = PyString_FromString("xlist");
	xinfo->uri = PyString_FromString("http://satine.sourceforge.net/schemas/kernel");
	xinfo->prefix = PyString_FromString("satine");

	xinfo->attrs_types = Py_None;
	xinfo->default_attrs = empty_dict;
	xinfo->required_attrs = empty_list;
	xinfo->none_attrs = empty_list;
	xinfo->items_dfa = Py_None;

	return (PyObject*) xinfo;
}

PyObject* 
PyXListInfo_New(PyTypeObject* type) {
	PyXListInfo* tpinfo;
	tpinfo =  PyObject_NEW(PyXListInfo, &PyXListInfo_Type);

	if (xlistinfo_create_tag(tpinfo, type) == -1)
		return NULL;
	if (xlistinfo_create_uriAndPrefix(tpinfo, type) == -1)
		return NULL;
	return (PyObject*) tpinfo;
}

int 
PyXListInfo_InitValidation(PyXListInfo* self, PyTypeObject* type) {
	Satine_WriteMoreInfo("xlistinfo init %s\n", type->tp_name);

	if (xlistinfo_compile_attrs(self, type) == -1)
		return -1;
	if (xlistinfo_compile_items(self, type) == -1) 
		return -1;

	return 0;
}

PyXListInfo*
PyXListInfo_FromType(PyTypeObject* type) {
	PyXListInfo* tpinfo;

	Satine_WriteMoreInfo("typeinfo for type %s\n", type->tp_name);
	tpinfo = (PyXListInfo*) PyDict_GetItemString(type->tp_dict, "tpinfo");
	if (tpinfo)
		return tpinfo;

	Satine_WriteMoreInfo("creation of new typeinfo for type %s\n", type->tp_name);
	tpinfo = (PyXListInfo*) PyXListInfo_New(type);
	if (tpinfo == NULL)
		return NULL;
	else
		PyDict_SetItemString(type->tp_dict, "tpinfo", (PyObject*) tpinfo);

	if (PyXListInfo_InitValidation(tpinfo, type) == -1) {
		Satine_WriteInfo("error during xlistinfo for type %s\n", type->tp_name);
		return NULL;
	}
	else
		return tpinfo;
}

static void
xlistinfo_dealloc(PyXListInfo *xinfo)
{
	Py_DECREF(xinfo->attrs_types);
	Py_DECREF(xinfo->default_attrs);
	Py_DECREF(xinfo->required_attrs);
	Py_DECREF(xinfo->none_attrs);
	Py_DECREF(xinfo->items_dfa);

	PyObject_DEL(xinfo);
}

static PyObject *
xlistinfo_tag_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->tag);
	return xinfo->tag;
}

static PyObject *
xlistinfo_uri_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->uri);
	return xinfo->uri;
}

static PyObject *
xlistinfo_prefix_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->prefix);
	return xinfo->prefix;
}

static PyObject *
xlistinfo_attrs_types_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->attrs_types);
	return xinfo->attrs_types;
}

static PyObject *
xlistinfo_default_attrs_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->default_attrs);
	return xinfo->default_attrs;
}

static PyObject *
xlistinfo_required_attrs_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->required_attrs);
	return xinfo->required_attrs;
}

static PyObject *
xlistinfo_none_attrs_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->none_attrs);
	return xinfo->none_attrs;
}

static PyObject *
xlistinfo_items_dfa_get(PyXListInfo *xinfo)
{
	Py_INCREF(xinfo->items_dfa);
	return xinfo->items_dfa;
}


static PyGetSetDef xlistinfo_getsets[] = {
	{"tag", (getter)xlistinfo_tag_get, NULL,
	 "the tag"},
	{"uri", (getter)xlistinfo_uri_get, NULL,
	 "the uri"},
	{"prefix", (getter)xlistinfo_prefix_get, NULL,
	 "the default prefix"},
	{"attrs_types", (getter)xlistinfo_attrs_types_get, NULL,
	 "the XML datatype for each attribute"},
	{"default_attrs", (getter)xlistinfo_default_attrs_get, NULL,
	 "the attributes that have a default value and that value"},
	{"required_attrs", (getter)xlistinfo_required_attrs_get, NULL,
	 "the list of attributes that are supposed to be assigned"},
	{"none_attrs", (getter)xlistinfo_none_attrs_get, NULL,
	 "the list of attributes without a default value"},
	{"items_dfa", (getter)xlistinfo_items_dfa_get, NULL,
	 "the DFA that describes allowed items"},
	{0}
};

static char xlistinfo_doc[] =
"information about the xlist class";

PyTypeObject PyXListInfo_Type = {
	PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,					/* ob_size */
	"xlistinfo",		/* tp_name */
	sizeof(PyXListInfo),	/* tp_basicsize */
	0,					/* tp_itemsize */
	/* methods */
	(destructor)xlistinfo_dealloc, 	/* tp_dealloc */
	0,					/* tp_print */
	0,					/* tp_getattr */
	0,					/* tp_setattr */
	0,					/* tp_compare */
	0,					/* tp_repr */
	0,					/* tp_as_number */
	0,					/* tp_as_sequence */
	0,					/* tp_as_mapping */
	0,					/* tp_hash */
	0,					/* tp_call */
	0,					/* tp_str */
	0,					/* tp_getattro */
	0,					/* tp_setattro */
	0,					/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	/* tp_flags */
 	xlistinfo_doc,		/* tp_doc */
 	0,					/* tp_traverse */
 	0,					/* tp_clear */
	0,					/* tp_richcompare */
	0,					/* tp_weaklistoffset */
	0,					/* tp_iter */
	0,					/* tp_iternext */
	0,					/* tp_methods */
	0,					/* tp_members */
	xlistinfo_getsets,	/* tp_getset */
	0,					/* tp_base */
	0,					/* tp_dict */
	0,					/* tp_descr_get */
	0,					/* tp_descr_set */
};

void initxlistinfo() {
	PyXListInfo_Type.ob_type = &PyType_Type;
	if (PyType_Ready(&PyXListInfo_Type) < 0)
		return;
	empty_list = PyList_New(0);
	empty_dict = PyDict_New();
}
